:- module(typer, [infr_types/0] ).

:- include(operators).

%

:- use_module(symtab,     [get_rules/2, set_object_property/3, get_object_property/3,
	                   delete_property/2, replace_object_property/3] ).

:- use_module(library(aggregates),   [findall/3] ).
:- use_module(library(lists),        [append/3, length/2] ).
:- use_module(library(terms_check),  [variant/2] ).
:- use_module(library(write),        [numbervars/3] ).

%

:- dynamic(opp/3).
:- multifile(opp/3).

% Comprobacion/Inferencia de tipos para Curry

is_constructor(C) :-
	get_object_property(C, constructs, _Type).

is_function(F) :-
	get_rules(F, _).

head_type(Head,Type,Env):-
	Head=id(_)@Pattern,
	head_type_pattern(Pattern,Type,Env).

head_type_pattern([],[],_):- !.

head_type_pattern([A],At,Env) :-
	!,
	term_type(A,At,Env).

head_type_pattern([C|Cs],(A->B),Env) :-
	term_type(C,A,Env),
	head_type_pattern(Cs,B,Env).

term_type(var('_'),_,_) :- 
	!.
term_type(var(V),VarType,Env) :-
	member(V-VarType,Env),
	!.

term_type(id(unit),tuple([_]),_) :-
	!.

term_type(id(C),Type,_) :-
	!,
	(
	    is_constructor(C),
	    get_object_property(C, type, Type)
	->
      	    true
	;
	    type_error('Undefined constructor %w',[C])
	).

term_type((C1,Cs),tuple([T1|Ts]),Env):-!,
	term_type(C1,T1,Env),
	(
	    Cs=..[','|_]
	->
	    term_type(Cs,tuple(Ts),Env) 
	; 
	    body_type(Cs,Cs_type,Env),
	    Ts = [Cs_type]
	).

term_type(id(C)@Args,Type,Env) :- 
	(
	    \+ is_constructor(C)
	->
	    type_error('Undefined symbol',[C])
	;
	    get_object_property(C, type, Ctype),
	    (
		check_cons_args(Ctype,Args,Env)  
	    ->
	        convert_func_list(Ctype,CtypeList),
		length(CtypeList,CtypeLength),
		length(Args,ArgsLength),
		ArgsLength<CtypeLength,
		length(A2,ArgsLength),
		append(A2,TypeList,CtypeList),
		convert_list_func(TypeList,Type)
	    ;
		true
	    )
	).

term_type([],list(_),_):-!.
term_type([T|Ts],list(Type),Env):-!,
	term_type(T,Type,Env),
	term_list_types(Ts,Types,Env),
	(member(X,Types),\+(X=Type) -> type_error('Different types in list',[]) ; true).
term_type(num(N),NumType,_):-
	(integer(N) -> NumType='Int' ; NumType='Float').
term_type(char(_),'Char',_).
term_type(str(_),list('Char'),_).

term_list_types([],[],_).
term_list_types([T|Ts],[Type|Types],Env):-
	term_type(T,Type,Env),
	term_list_types(Ts,Types,Env).

convert_func_list(Type,[A|Blist]):-
	\+var(Type),Type=(A->B),!,
	convert_func_list(B,Blist).
convert_func_list(A,[A]).

convert_list_func([A],A):-!.
convert_list_func([A|As],(A->B)):-
	convert_list_func(As,B).

check_cons_args((A->B),[Aarg|Bs],Env):-!,
	term_type(Aarg,AargType,Env),
	(unify(A,AargType)-> true ;
	    type_error('Type mismatch: %w should be %w',[Aarg,A])),
	check_cons_args(B,Bs,Env).

check_cons_args((_->_),[],_) :-
	!.
check_cons_args(_,[],_) :-
	!.
check_cons_args(A,[Aarg],Env) :-
	term_type(Aarg,AargType,Env),
	(
	    unify(A, AargType) -> 
	    true 
	;
	    type_error('Type mismatch: w should be w',[Aarg,A])
	).

unify(X,Y) :-
	X = Y,
	!.

unify(X,Y) :- 
	(
	    X = var(_)
	;
	    Y = var(_)
	).

unify_list([],_):-!.
unify_list(_,[]):-!.
unify_list([A|As],[B|Bs]):-
	unify(A,B),unify_list(As,Bs).

%%%%%%%%%%% Calculo de tipo del cuerpo: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

body_type(A,_,_):-
	var(A),
	!.

body_type((E1,Es),tuple([T1|Ts]),Env):-!,
	body_type(E1,T1,Env),
	(
	    Es=..[','|_]
	->
	    body_type(Es, tuple(Ts), Env) 
	; 
	    body_type(Es,Es_type,Env),Ts=[Es_type]
	).

body_type(num(N),NumType,_):- 
	!,
	(
	    integer(N)
	->
	NumType='Int' 
	;
	    NumType= 'Float'
	).

body_type(id(unit),tuple([_]),_) :-
	!.
body_type(char(_),'Char',_):- 
	!.
body_type(str(_),list('Char'),_):-!.
body_type(cond(Constraint,Exp),Type,Env):-!,
	body_type(Constraint,'Constraint',Env),
	body_type(Exp,Type,Env).

body_type([],list(_),_):-!.

body_type([E|Es],list(Type),Env):-!,
	body_type(E,Type,Env),
	body_type_list(Es,Types,Env),
	(
	    member(X,Types),
	    \+(X=Type) 
	-> 
	    type_error('Different types in list',[]) 
	; 
	    true
	).

body_type(constraint(List),'Constraint',Env):-!,
	body_type_list(List,Types,Env),
	(
	    member(X,Types),
	    \+(X='Constraint')
	->
	    type_error('A Constraint type was expected instead of %w',[X])
	;
	    true
	).

body_type(id(C),Type,_):-
	is_constructor(C),
	!,
	(
	    get_object_property(C,type,Type)
	->
 	    true
	;
	    type_error('Undefined symbol %w',[C])
	).

body_type(id(F),Type,_):-
	is_function(F),
	!,
	(
	    get_object_property(F, type, Type)
	->
	    true 
	; 
	    true
	).

body_type(id(C)@Args,Type,Env):-
	is_constructor(C),
	!,
	(
	    \+ get_object_property(C, type, Ctype) 
	->
	    type_error('Undefined symbol %w',[C])
	;
	    (
		get_object_property(C, type, Ctype),
		check_args_type(Ctype,Args,Env) 
	    ->
	        convert_func_list(Ctype,CtypeList),
		length(CtypeList,CtypeLength),
		length(Args,ArgsLength),
		ArgsLength<CtypeLength,
		length(A2,ArgsLength),
		append(A2,TypeList,CtypeList),
		convert_list_func(TypeList,Type)
	    ;
		true)
	).

body_type(id(F)@Args,Type,Env):-
	is_function(F),
	!,
	(
	    get_object_property(F, type, Ftype) 
	-> 
	    true 
	;
	    get_object_property(F, arity, Parity),
	    length(List,Parity),
	    append(List,[_],List2),
	    convert_list_func(List2,Ftype)
	),      
	(
	    check_args_type(Ftype,Args,Env) 
	->
	    convert_func_list(Ftype,FtypeList),
	    length(FtypeList,FtypeLength),
	    length(Args,ArgsLength),
	    ArgsLength<FtypeLength,
	    length(A2,ArgsLength),
	    append(A2,TypeList,FtypeList),
	    convert_list_func(TypeList,Type)
						%      update(F,infrType,Ftype)
	;
	    true
	).


body_type(BinExp,Type,Env):-
	BinExp=..[BOp|LeftRight],
	(
	    BOp='.'
	->
	    BinOp=':' 
	; 
	    BinOp=BOp
	),
	opp(_,_,BinOp),
	!,        
	(
	    \+ get_object_property(BinOp, type, Otype)
	->
	    append(LeftRight,Tl,[_,_,_]),
	    convert_list_func(Tl,Type)
	;
	    get_object_property(BinOp, type, Otype),
	    (
		check_args_type(Otype,LeftRight,Env) 
	    ->
	        convert_func_list(Otype,OtypeList),
		append(_,[Type],OtypeList)
	    ;
		true
	    )
	).

body_type(var('_'),_,_):-!.
body_type(var(V),Vtype,Env):-
	member(V-Vtype,Env),!.

body_type(var(V)@Args,FinalType,Env):-
	member(V-Vtype,Env),!,
	body_type_list(Args,ArgsTypes,Env),
	append(ArgsTypes,[FinalType],ArgsTypes2),
						%  convert_func_list(Vtype,VtypeList),
						%  unify_list(ArgsTypes2,VtypeList),
	convert_list_func(ArgsTypes2,Vtype2),
	(Vtype=Vtype2->true;type_error('Type mismatch: %w expected instead of %w',[Vtype,Vtype2])).
    
body_type_list([],[],_).
body_type_list([A|As],[T|Ts],Env):-
	body_type(A,T,Env),
	body_type_list(As,Ts,Env).
   

check_args_type(Type,[Aarg|Bs],Env):-
	\+var(Type),
	Type=(A->B),
	body_type(Aarg,AargType,Env),
	(unify(A,AargType)->true ;
	    type_error('Type mismatch  (%w should be %w)',[AargType,A])),
	    check_args_type(B,Bs,Env).

check_args_type(Type,[],_):-
	\+var(Type),Type=(_->_).

check_args_type(_,[],_):-!.
check_args_type(A,[Aarg],Env):-
	body_type(Aarg,AargType,Env),
	(unify(A,AargType)-> true ;
	    type_error('Type mismatch (%w should be %w)',[AargType,A])).

choice_list_type([],TypeAcum,TypeAcum,_).
choice_list_type([(constraint(C)->Exp)|Ls],TypeAcum,Acum,Env):-
	(body_type(constraint(C),'Constraint',Env)->true
   ;
	type_error('Invalid Constraint in choice expression',[])),
	body_type(Exp,BType,Env),
	(TypeAcum=BType->true
    ; 
	type_error('Expression in choice are not the same type',[])),
	choice_list_type(Ls,TypeAcum,Acum,Env).

%%%%%%%% Tipo de una regla %%%%%%%%%%%%%%%%%%%%%%%%%%

rule_type((id(F)@Args):=Body,Type,Env):-
	head_type(id(F)@Args,Th,Env),
	body_type(Body,Tb,Env),
	convert_func_list(Th,ThList),
	convert_func_list(Tb,TbList),
	(Th==[] -> TypeList=TbList ; append(ThList,TbList,TypeList)),
	convert_list_func(TypeList,Type).  

%%%%% Tipo de la funcion F: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

fun_type(F,Type):-
	set_object_property('@Type_data', current_func, F),
	get_rules(F, Reglas),
	(
	    Reglas = [R1|Rs]
	->
	    set_object_property('@Type_data', current_rule, 1),
	    rule_type(R1,TypeAux,_Env), 
	    set_object_property(F, type, TypeAux),
	    fun_type_rule_list(F,Type,Rs,2) 
	; 
	    true
	).

fun_type_rule_list(F,Type,[],_):-
	get_object_property(F, type, Type),
	!.

fun_type_rule_list(F, _Type, [(id(F)@Head):=Body|Reglas], NumRule) :-
	set_object_property('@Type_data', current_rule, NumRule),
	get_object_property(F, type, InfType),
	!,
	convert_func_list(InfType,InfTypeList),
	append(HeadList,[BodyTypeInf],InfTypeList),
	convert_list_func(HeadList,HeadTypeInf),
	head_type(id(F)@Head,HeadType,Env),
	body_type(Body,BodyType,Env),
	(
	    HeadTypeInf = HeadType,
	    BodyTypeInf = BodyType
	->
	    true 
	; 
	    type_error('Rules have different types for %w',[F])
	),
	convert_func_list(HeadTypeInf,HTIL),
	convert_func_list(BodyTypeInf,BTIL),
	append(HTIL,BTIL,ITL),
	convert_list_func(ITL,InferredType),
	set_object_property(F, type, InferredType),
	N1 is NumRule+1,
	fun_type_rule_list(F,_Type,Reglas,N1).

%%%%%%%%%%%%%% Dar una pasada a las funciones: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear_changes:-
	delete_property(type, change).

pass_types:-
	findall(F, (get_rules(F, R), \+ R = [builtin]), Fs),
	get_object_property('@Type_data', cycle, C),
	length(Fs, L),
	(
	    get_object_property(type, change, FChange)
	->
	    true
	;
	    true
	),
	(
	    C > L 
	-> 
	    set_object_property('@Type_data', current_func, FChange),
	    set_object_property('@Type_data', current_rule, 1),
	    clear_changes,
	    type_error('Infinite type for function %w',[FChange])
	;
	    clear_changes,
	    fun_type_list(Fs)
	). 

fun_type_list([]).
fun_type_list([F|Fs]):-
	(
	    member(F, [return,done,getChar,putChar, '@Objective']) 
	-> 
	    true 
	;
	    (
		get_object_property(F, type, InfFType) 
	    ->
		true 
	    ; 
		true
	    ),
	    fun_type(F, FType),
	    (
		% This is commented out for the moment.
		true
		%variant(FType, InfFType)
	    ->
	        true 
	    ; 
		set_object_property(type, change, F)
	    )
	),
	fun_type_list(Fs).

% Necesita esto, pero debera buscar en las tablas propais.
% opp(1,fx,'not').

opp(1,fx,'-').	
opp(2,xfx,mod).
opp(2,yfx,div).
opp(2,yfx,'*' ). 
opp(2,yfx,'/' ).
opp(3,xfy,':' ).
opp(3,yfx,'-' ).
opp(3,yfx,'+' ). 
opp(4,xfx,'=' ).
opp(4,xfx,'==').
opp(4,xfx,'<' ).
opp(4,xfx,'=<').
opp(4,xfx,'>' ).
opp(4,xfx,'>=').
opp(4,xfx,'<>').
opp(4,xfy,'>>').
opp(4,xfy,'>>=').
opp(5,xfy,'&&' ). 
opp(6,xfy,'||' ). 

% Dar pasadas hasta que se encuentre el punto fijo de los tipos:

infr_types:- % Predicado principal para inferir tipos
	refresh_var_types,
	set_object_property('@Type_data',cycle,0),
	types.

types:- 
	pass_types,
	get_object_property(type, change, _),
	!,
	get_object_property('@Type_data', cycle, C),
	C1 is C+1,
	set_object_property('@Type_data', cycle, C1),
	types.
types:-
	set_object_property('@Type_data', current_func, 'Goal'),
	set_object_property('@Type_data', current_rule, 1),
	(
	    get_rules('@Objective', [ObjTree]),
	    body_type(ObjTree,_,_)
	->
	    true
	;
	    true
	).

%%%%%%%%% Comprobar que el tipo inferido concuerda con el especificado:

check_types:-
	delete_property('@Type_data', _),
	findall(F, (get_rules(F, R), \+ R = [builtin]), Fs),
	check_types_list(Fs).

check_types_list([]):-
	delete_property('@Type_data', _).

check_types_list([_|Fs]):-
	check_types_list(Fs).    


%%%%%%%%%%%%%%%% Cambiar las variables de tipo por var. libres:

refresh_var_types:-
	findall(F, (is_function(F), get_object_property(F, type, _)), Fn),
	findall(C, get_object_property(C, constructs, _Type), Constructors1),
	append(Fn, Constructors1, Constructors),
	refresh_var_types1(Constructors).

refresh_var_types1([]).
refresh_var_types1([C|Cs]):-
	get_object_property(C, type, CType),
	convert_func_list(CType,CList),
	append(_,[CFinal],CList),!,
	% If final type is a var, we cannot guarantee that =.. will unify.
        ( 
	    var(CFinal) ->
	    TypeVars = []
	;
	    CFinal =..[_|TypeVars]
	),
	refresh_var_type_exp(CType,TypeVars,CtypeVars,_Env),
	replace_object_property(C, type, CtypeVars),
	refresh_var_types1(Cs).

refresh_var_type_exp(Exp,ShapeVars,ExpVars,Env):-
	\+var(Exp),
	Exp=..[Op|List],List=[_|_],!,
	refresh_var_type_exp_list(List,ShapeVars,ListVars,Env),
	ExpVars=..[Op|ListVars].
refresh_var_type_exp(Id,ShapeVars,IdVar,Env):-
	(
	    member(Id,ShapeVars)
	->
	    member(Id-IdVar,Env) 
	; 
	    IdVar=Id
	),
	!.

refresh_var_type_exp_list([],_,[],_).
refresh_var_type_exp_list([E|Es],ShapeVars,[EVars|EsVars],Env):-
	refresh_var_type_exp(E,ShapeVars,EVars,Env),
	refresh_var_type_exp_list(Es,ShapeVars,EsVars,Env).

%%%%%%%%%%%%% Error de tipo: %%%%%%%%%%%%%%%%%%%%%%%%

type_error(Format, Vars) :-
	throw(type_error(Format, Vars)).
% 	numbervars(Vars,0,_),
% 	(lookup('@Type_data',current_func,CFunc),
% 	lookup('@Type_data',current_rule,CRule) 
%     ->
% 	swritef(ErrorPrefix,"In function %w, rule %w: ",[CFunc,CRule])
%     ;
% 	name(ErrorPrefix,[])),
% 	swritef(ErrorSufix,Format,Vars),
% 	name(ErrorPrefix,ErrorPrefixList),
% 	name(ErrorSufix,ErrorSufixList),
% 	append(ErrorPrefixList,ErrorSufixList,ErrorList),
% 	name(Error,ErrorList),
% 	(lookup('@Type_Error',_,Error)->true ;
% 	    findall(E,lookup('@Type_Error',E,_),Es),
% 	    length(Es,N),
% 	    N1 is N+1,
% 	    (lookup('@Type_data',current_func,CFunc),
% 	    lookup('@Type_data',current_rule,CRule) 
% 	->
% 	update('@Type_Error',error(N1),Error)
%     ; true)).



%%%%%%%%%% Prueba: %%%%%%%%%%%%%%

% prueba_t(T,Tb,E):-
% 	parse('a'),
% 	lookup(_,abstree,Regla),
% 	Regla=(Head:=Body),
% 	head_type(Head,T,E),
% 	body_type(Body,Tb,E).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

